/* Iain Murray, 2008 */

#include "mex.h"
#include <math.h>

/* This code uses ideas from:
 * http://graphics.stanford.edu/~seander/bithacks.html#IntegerLogIEEE64Float */

/* If you get an error inside this ifndef, define EXP_PART elsewhere.
 * For example, pass your compiler the following option:
 * -DEXP_PART=$(python -c "import sys; print(0 if sys.byteorder=='big' else 1)")
 */
#ifndef EXP_PART
#include <endian.h>
#define EXP_PART (__FLOAT_WORD_ORDER==LITTLE_ENDIAN)
#endif

#ifndef min
#define min(x,y) (((x) > (y)) ? (y) : (x))
#endif
#ifndef INFINITY
#define INFINITY HUGE_VAL
#endif
#ifndef NAN
#define NAN (INFINITY - INFINITY)
#endif

int naive_sumlog(double x[], int size, double* answer)
{
    int i;
    double sum = 0;
    for (i = 0; i < size; ++i) {
        sum += log(x[i]);
    }
    *answer = sum;
    return 0;
}


int sumlog(double x[], int size, double* answer)
/* Gets each item in x as m * 2^e where m is between 1 and 2.
 * Multiplies all the m's and adds all the e's to find log(prod(x)) without
 * over- or under-flowing. */
{
    int exponent, i;
    int start = 0;
    int end;
    int type;
    int nan_flag = 0;
    int denormal_nan_inf_neg_or_zero;
    union { unsigned int u[2]; double d; } tmp;
    double prod = 1.0;
    double sum = 0.0;
    const int BATCH_SIZE = 1023; /* (2-eps)^1024 doesn't overflow */

    while (1) {
        end = min(start + BATCH_SIZE, size);
        /* Main Loop */
        for (i = start; i < end; ++i) {
            tmp.d = x[i];
            exponent = (tmp.u[EXP_PART] >> 20) - 1023;
            denormal_nan_inf_neg_or_zero = ((exponent == -1023) || (exponent > 1023));
            if (denormal_nan_inf_neg_or_zero)
                break;
            tmp.u[EXP_PART] &= 0x800FFFFF; /* 0b10000000000011111111111111111111 */
            tmp.u[EXP_PART] |= 0x3FF00000; /* 0b00111111111100000000000000000000 */
            prod *= tmp.d;
            sum += exponent;
        }
        
        /* Check if loop broke out because of special cases and deal with them.
         * This isn't optimized. For maximum performance, one should ensure that
         * all of the inputs to sumlog() are nice. */
        if (i < end) {
            if (tmp.d < 0) {
                *answer = NAN;
                return -1;
            }
            type = fpclassify(tmp.d);
            switch (type) {
                case FP_ZERO:
                    prod = 0.0;
                    break;
                case FP_INFINITE:
                    sum = INFINITY;
                    break;
                case FP_NAN:
                    nan_flag = 1;
                    break;
                case FP_SUBNORMAL:
                    /* denormalized numbers...YUCK! */
                    tmp.d *= (1 << 30);
                    tmp.d *= (1 << 30);
                    sum -= 60;
                    exponent = (tmp.u[EXP_PART] >> 20) - 1023;
                    tmp.u[EXP_PART] &= 0x800FFFFF; /* 0b10000000000011111111111111111111 */
                    tmp.u[EXP_PART] |= 0x3FF00000; /* 0b00111111111100000000000000000000 */
                    prod *= tmp.d;
                    sum += exponent;
                    break;
                case FP_NORMAL:
                    return -2;
                    break;
                default:
                    return -3;
            }
            /* Correct where we actually got to for next bit of code. */
            end = i + 1;
        }

        /* Stop prod from over-flowing every so often */
        if (end < size) {
            if (prod > 0) {
                tmp.d = prod;
                sum += (tmp.u[EXP_PART] >> 20) - 1023;
                tmp.u[EXP_PART] &= 0x800FFFFF; /* 0b10000000000011111111111111111111 */
                tmp.u[EXP_PART] |= 0x3FF00000; /* 0b00111111111100000000000000000000 */
                prod = tmp.d;
            }
            start = end;
        } else {
            break;
        }
    }
    if (nan_flag)
        *answer = NAN;
    else
        *answer = log(prod) + sum*M_LN2;
    return 0;
}

int sumlogf(float x[], int size, float* answer)
/* Gets each item in x as m * 2^e where m is between 1 and 2.
 * Multiplies all the m's and adds all the e's to find log(prod(x)) without
 * over- or under-flowing. */
{
    int exponent, i;
    int start = 0;
    int end;
    int type;
    int nan_flag = 0;
    int denormal_nan_inf_neg_or_zero;
    union { unsigned int u; float f; } tmp;
    float prod = 1.0;
    float sum = 0.0;
    const int BATCH_SIZE = 127; /* (2-eps)^128 doesn't overflow (for singles) */
    const int FLOAT_EXP_BIAS = 127;

    while (1) {
        end = min(start + BATCH_SIZE, size);
        /* Main Loop */
        for (i = start; i < end; ++i) {
            tmp.f = x[i];
            exponent = (tmp.u >> 23) - FLOAT_EXP_BIAS;
            denormal_nan_inf_neg_or_zero = ((exponent == -FLOAT_EXP_BIAS) || (exponent > FLOAT_EXP_BIAS));
            if (denormal_nan_inf_neg_or_zero)
                break;
            tmp.u &= 0x807FFFFF; /* 0b10000000011111111111111111111111 */
            tmp.u |= 0x3F800000; /* 0b00111111100000000000000000000000 */
            prod *= tmp.f;
            sum += exponent;
        }
        
        /* Check if loop broke out because of special cases and deal with them.
         * This isn't optimized. For maximum performance, one should ensure that
         * all of the inputs to sumlog() are nice. */
        if (i < end) {
            if (tmp.f < 0) {
                *answer = NAN;
                return -1;
            }
            type = fpclassify(tmp.f);
            switch (type) {
                case FP_ZERO:
                    prod = 0.0;
                    break;
                case FP_INFINITE:
                    sum = INFINITY;
                    break;
                case FP_NAN:
                    nan_flag = 1;
                    break;
                case FP_SUBNORMAL:
                    /* denormalized numbers...YUCK! */
                    tmp.f *= (1 << 30);
                    sum -= 30;
                    exponent = (tmp.u >> 23) - FLOAT_EXP_BIAS;
                    tmp.u &= 0x807FFFFF; /* 0b10000000011111111111111111111111 */
                    tmp.u |= 0x3F800000; /* 0b00111111100000000000000000000000 */
                    prod *= tmp.f;
                    sum += exponent;
                    break;
                case FP_NORMAL:
                    return -2;
                    break;
                default:
                    return -3;
            }
            /* Correct where we actually got to for next bit of code. */
            end = i + 1;
        }

        /* Stop prod from over-flowing every so often */
        if (end < size) {
            if (prod > 0) {
                tmp.f = prod;
                sum += (tmp.u >> 23) - FLOAT_EXP_BIAS;
                tmp.u &= 0x807FFFFF; /* 0b10000000011111111111111111111111 */
                tmp.u |= 0x3F800000; /* 0b00111111100000000000000000000000 */
                prod = tmp.f;
            }
            start = end;
        } else {
            break;
        }
    }
    if (nan_flag)
        *answer = NAN;
    else
        *answer = logf(prod) + sum*M_LN2;
    return 0;
}



void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    const mxArray *mxIn;
    size_t i, M, N;
    int status = 0;
    mxClassID class;
    int (*sumlog_fn) (double*, int, double*);
    double *outArray, *inArray;
    float *outArrayf, *inArrayf;

    if (nrhs!=1)
        mexErrMsgTxt("There must be one and only input.");
    if (nlhs>1) /* can be "zero", in which case Matlab allows one */
        mexErrMsgTxt("There must be one and only output.");

    /* get information about the input */
    mxIn = prhs[0];
    if (mxIsSparse(mxIn))
        mexErrMsgTxt("Sparse matrices are not supported.");
    class = mxGetClassID(mxIn);
    if ((class != mxDOUBLE_CLASS) && (class != mxSINGLE_CLASS))
        mexErrMsgTxt("Only double or single precision full matrices are supported.");
    if (mxGetNumberOfDimensions(mxIn) != 2)
        mexErrMsgTxt("Input must be a matrix.");
    M = mxGetM(mxIn);
    N = mxGetN(mxIn);

    if (class == mxDOUBLE_CLASS) {
        inArray = mxGetPr(mxIn);
        plhs[0] = mxCreateDoubleMatrix(1, N, 0);
        outArray = mxGetPr(plhs[0]);
        for (i = 0; i < N; ++i) {
            status = sumlog(inArray, M, &outArray[i]);
            if (status)
                break;
            inArray += M;
        }
    } else {
        inArrayf = mxGetData(mxIn);
        plhs[0] = mxCreateNumericMatrix(1, N, mxSINGLE_CLASS, 0);
        outArrayf = mxGetData(plhs[0]);
        for (i = 0; i < N; ++i) {
            status = sumlogf(inArrayf, M, &outArrayf[i]);
            if (status)
                break;
            inArrayf += M;
        }
    }

    switch (status) {
        case 0:
            break;
        case -1:
            mexErrMsgTxt("Can't find real log of a negative value.");
            break;
        case -3:
            mexErrMsgTxt("Unknown floating point type encountered.");
            break;
        case -2:
            /* fall through */
        default:
            mexErrMsgTxt("Code reached unexpected outcome.");
    }
}

